mod client;
mod messages;
mod prelude;
mod server;

use prelude::*;

struct Args {
    cmd: Subcommand,
}

enum Subcommand {
    Client,
    Server,
}

impl Args {
    fn new() -> Result<Self> {
        let mut args = std::env::args().skip(1);
        let cmd = match args.next().as_deref() {
            None => bail!("Need a subcommand, e.g. `client` or `server`"),
            Some("client") => Subcommand::Client,
            Some("server") => Subcommand::Server,
            Some(_) => bail!("Invalid subcommand"),
        };
        Ok(Self { cmd })
    }
}

#[tokio::main]
async fn main() -> Result<()> {
    let args = Args::new()?;
    tracing_subscriber::fmt::init();

    match args.cmd {
        Subcommand::Client => run_client().await,
        Subcommand::Server => run_server().await,
    }
}

async fn run_client() -> Result<()> {
    let mut ctrl_c = signal(SignalKind::interrupt())?;
    let mut app = client::App::new().await?;
    loop {
        let ctl = poll_fn(|cx| {
            tracing::debug!("Loop");
            if ctrl_c.poll_recv(cx).is_ready() {
                tracing::debug!("Caught Ctrl-C");
                return Poll::Ready(ControlFlow::Break(Ok(())));
            }

            match app.poll_run(cx) {
                Poll::Pending => Poll::Pending,
                Poll::Ready(Ok(())) => Poll::Ready(ControlFlow::Continue(())),
                Poll::Ready(Err(err)) => Poll::Ready(ControlFlow::Break(Err(err))),
            }
        })
        .await;
        if let ControlFlow::Break(result) = ctl {
            return result;
        }
    }
}

async fn run_server() -> Result<()> {
    let mut ctrl_c = signal(SignalKind::interrupt())?;
    let mut app = server::App::new().await?;
    poll_fn(|cx| {
        tracing::debug!("Loop");
        if ctrl_c.poll_recv(cx).is_ready() {
            tracing::debug!("Caught Ctrl-C");
            return Poll::Ready(Ok(()));
        }

        app.poll_run(cx)
    })
    .await
}
